@@ -1,5 +1,5 @@
 /*
- * $Id: ossl_cipher.c 23204 2009-04-18 14:10:06Z akr $
+ * $Id: ossl_cipher.c 32724 2011-07-28 13:52:57Z nahi $
  * 'OpenSSL for Ruby' project
  * Copyright (C) 2001-2002  Michal Rokos <m.rokos@sh.cvut.cz>
  * All rights reserved.
@@ -10,17 +10,24 @@
  */
 #include "ossl.h"
 
+#define WrapCipher(obj, klass, ctx) \
+    (obj) = Data_Wrap_Struct((klass), 0, ossl_cipher_free, (ctx))
 #define MakeCipher(obj, klass, ctx) \
-    obj = Data_Make_Struct(klass, EVP_CIPHER_CTX, 0, ossl_cipher_free, ctx)
+    (obj) = Data_Make_Struct((klass), EVP_CIPHER_CTX, 0, ossl_cipher_free, (ctx))
+#define AllocCipher(obj, ctx) \
+    memset(DATA_PTR(obj) = (ctx) = ALLOC(EVP_CIPHER_CTX), 0, sizeof(EVP_CIPHER_CTX))
+#define GetCipherInit(obj, ctx) do { \
+    Data_Get_Struct((obj), EVP_CIPHER_CTX, (ctx)); \
+} while (0)
 #define GetCipher(obj, ctx) do { \
-    Data_Get_Struct(obj, EVP_CIPHER_CTX, ctx); \
-    if (!ctx) { \
+    GetCipherInit((obj), (ctx)); \
+    if (!(ctx)) { \
 	ossl_raise(rb_eRuntimeError, "Cipher not inititalized!"); \
     } \
 } while (0)
 #define SafeGetCipher(obj, ctx) do { \
-    OSSL_Check_Kind(obj, cCipher); \
-    GetCipher(obj, ctx); \
+    OSSL_Check_Kind((obj), cCipher); \
+    GetCipher((obj), (ctx)); \
 } while (0)
 
 /*
@@ -51,7 +58,7 @@ ossl_cipher_new(const EVP_CIPHER *cipher)
     EVP_CIPHER_CTX *ctx;
 
     ret = ossl_cipher_alloc(cCipher);
-    GetCipher(ret, ctx);
+    AllocCipher(ret, ctx);
     EVP_CIPHER_CTX_init(ctx);
     if (EVP_CipherInit_ex(ctx, cipher, NULL, NULL, NULL, -1) != 1)
 	ossl_raise(eCipherError, NULL);
@@ -74,12 +81,10 @@ ossl_cipher_free(EVP_CIPHER_CTX *ctx)
 static VALUE
 ossl_cipher_alloc(VALUE klass)
 {
-    EVP_CIPHER_CTX *ctx;
     VALUE obj;
 
-    MakeCipher(obj, klass, ctx);
-    EVP_CIPHER_CTX_init(ctx);
-	
+    WrapCipher(obj, klass, 0);
+
     return obj;
 }
 
@@ -97,13 +102,26 @@ ossl_cipher_initialize(VALUE self, VALUE str)
     EVP_CIPHER_CTX *ctx;
     const EVP_CIPHER *cipher;
     char *name;
+    unsigned char key[EVP_MAX_KEY_LENGTH];
 
     name = StringValuePtr(str);
-    GetCipher(self, ctx);
+    GetCipherInit(self, ctx);
+    if (ctx) {
+	ossl_raise(rb_eRuntimeError, "Cipher already inititalized!");
+    }
+    AllocCipher(self, ctx);
+    EVP_CIPHER_CTX_init(ctx);
     if (!(cipher = EVP_get_cipherbyname(name))) {
 	ossl_raise(rb_eRuntimeError, "unsupported cipher algorithm (%s)", name);
     }
-    if (EVP_CipherInit_ex(ctx, cipher, NULL, NULL, NULL, -1) != 1)
+    /*
+     * The EVP which has EVP_CIPH_RAND_KEY flag (such as DES3) allows
+     * uninitialized key, but other EVPs (such as AES) does not allow it.
+     * Calling EVP_CipherUpdate() without initializing key causes SEGV so we
+     * set the data filled with "\0" as the key by default.
+     */
+    memset(key, 0, EVP_MAX_KEY_LENGTH);
+    if (EVP_CipherInit_ex(ctx, cipher, NULL, key, NULL, -1) != 1)
 	ossl_raise(eCipherError, NULL);
 
     return self;
@@ -112,11 +130,14 @@ static VALUE
 ossl_cipher_copy(VALUE self, VALUE other)
 {
     EVP_CIPHER_CTX *ctx1, *ctx2;
-	
+
     rb_check_frozen(self);
     if (self == other) return self;
 
-    GetCipher(self, ctx1);
+    GetCipherInit(self, ctx1);
+    if (!ctx1) {
+	AllocCipher(self, ctx1);
+    }
     SafeGetCipher(other, ctx2);
     if (EVP_CIPHER_CTX_copy(ctx1, ctx2) != 1)
 	ossl_raise(eCipherError, NULL);
@@ -170,7 +191,7 @@ ossl_cipher_reset(VALUE self)
     GetCipher(self, ctx);
     if (EVP_CipherInit_ex(ctx, NULL, NULL, NULL, NULL, -1) != 1)
 	ossl_raise(eCipherError, NULL);
-		
+
     return self;
 }
 
@@ -189,7 +210,7 @@ ossl_cipher_init(int argc, VALUE *argv, VALUE self, int mode)
 	 * keeping this behaviour for backward compatibility.
 	 */
 	const char *cname  = rb_class2name(rb_obj_class(self));
-	rb_warn("argumtents for %s#encrypt and %s#decrypt were deprecated; "
+	rb_warn("arguments for %s#encrypt and %s#decrypt were deprecated; "
                 "use %s#pkcs5_keyivgen to derive key and IV",
                 cname, cname, cname);
 	StringValue(pass);
@@ -204,7 +225,7 @@ ossl_cipher_init(int argc, VALUE *argv, VALUE self, int mode)
 	    else memcpy(iv, RSTRING_PTR(init_v), sizeof(iv));
 	}
 	EVP_BytesToKey(EVP_CIPHER_CTX_cipher(ctx), EVP_md5(), iv,
-		       (unsigned char *)RSTRING_PTR(pass), RSTRING_LEN(pass), 1, key, NULL);
+		       (unsigned char *)RSTRING_PTR(pass), RSTRING_LENINT(pass), 1, key, NULL);
 	p_key = key;
 	p_iv = iv;
     }
@@ -280,14 +301,14 @@ ossl_cipher_pkcs5_keyivgen(int argc, VALUE *argv, VALUE self)
     if(!NIL_P(vsalt)){
 	StringValue(vsalt);
 	if(RSTRING_LEN(vsalt) != PKCS5_SALT_LEN)
-	    rb_raise(eCipherError, "salt must be an 8-octet string");
+	    ossl_raise(eCipherError, "salt must be an 8-octet string");
 	salt = (unsigned char *)RSTRING_PTR(vsalt);
     }
     iter = NIL_P(viter) ? 2048 : NUM2INT(viter);
     digest = NIL_P(vdigest) ? EVP_md5() : GetDigestPtr(vdigest);
     GetCipher(self, ctx);
     EVP_BytesToKey(EVP_CIPHER_CTX_cipher(ctx), digest, salt,
-		   (unsigned char *)RSTRING_PTR(vpass), RSTRING_LEN(vpass), iter, key, iv); 
+		   (unsigned char *)RSTRING_PTR(vpass), RSTRING_LENINT(vpass), iter, key, iv);
     if (EVP_CipherInit_ex(ctx, NULL, NULL, key, iv, -1) != 1)
 	ossl_raise(eCipherError, NULL);
     OPENSSL_cleanse(key, sizeof key);
@@ -305,7 +326,7 @@ ossl_cipher_pkcs5_keyivgen(int argc, VALUE *argv, VALUE self)
  *  +data+ is a nonempty string.
  *  +buffer+ is an optional string to store the result.
  */
-static VALUE 
+static VALUE
 ossl_cipher_update(int argc, VALUE *argv, VALUE self)
 {
     EVP_CIPHER_CTX *ctx;
@@ -317,8 +338,8 @@ ossl_cipher_update(int argc, VALUE *argv, VALUE self)
 
     StringValue(data);
     in = (unsigned char *)RSTRING_PTR(data);
-    if ((in_len = RSTRING_LEN(data)) == 0)
-        rb_raise(rb_eArgError, "data must not be empty");
+    if ((in_len = RSTRING_LENINT(data)) == 0)
+        ossl_raise(rb_eArgError, "data must not be empty");
     GetCipher(self, ctx);
     out_len = in_len+EVP_CIPHER_CTX_block_size(ctx);
 
@@ -345,7 +366,7 @@ ossl_cipher_update(int argc, VALUE *argv, VALUE self)
  *
  *  See EVP_CipherFinal_ex for further information.
  */
-static VALUE 
+static VALUE
 ossl_cipher_final(VALUE self)
 {
     EVP_CIPHER_CTX *ctx;
@@ -431,7 +452,7 @@ ossl_cipher_set_iv(VALUE self, VALUE iv)
 
 /*
  *  call-seq:
- *     cipher.key_length = integer -> integer
+ *     cipher.key_len = integer -> integer
  *
  *  Sets the key length of the cipher.  If the cipher is a fixed length cipher then attempting to set the key
  *  length to any value other than the fixed value is an error.
@@ -445,7 +466,7 @@ ossl_cipher_set_key_length(VALUE self, VALUE key_length)
 {
     int len = NUM2INT(key_length);
     EVP_CIPHER_CTX *ctx;
- 
+
     GetCipher(self, ctx);
     if (EVP_CIPHER_CTX_set_key_length(ctx, len) != 1)
         ossl_raise(eCipherError, NULL);
@@ -494,13 +515,13 @@ CIPHER_0ARG_INT(block_size)
 #if 0
 /*
  *  call-seq:
- *     cipher.key_length -> integer
+ *     cipher.key_len -> integer
  *
  */
 static VALUE ossl_cipher_key_length() { }
 /*
  *  call-seq:
- *     cipher.iv_length -> integer
+ *     cipher.iv_len -> integer
  *
  */
 static VALUE ossl_cipher_iv_length() { }
@@ -515,11 +536,11 @@ static VALUE ossl_cipher_block_size() { }
 /*
  * INIT
  */
-void 
+void
 Init_ossl_cipher(void)
 {
-#if 0 /* let rdoc know about mOSSL */
-    mOSSL = rb_define_module("OpenSSL");
+#if 0
+    mOSSL = rb_define_module("OpenSSL"); /* let rdoc know about mOSSL */
 #endif
     cCipher = rb_define_class_under(mOSSL, "Cipher", rb_cObject);
     eCipherError = rb_define_class_under(cCipher, "CipherError", eOSSLError);
